local MAJOR = "LibScriptableDisplayWidgetColor-1.0" 
local MINOR = 14

assert(LibStub, MAJOR.." requires LibStub") 
local WidgetColor = LibStub:NewLibrary(MAJOR, MINOR)
if not WidgetColor then return end
local LibWidget = LibStub("LibScriptableDisplayWidget-1.0", true)
assert(LibWidget, MAJOR .. " requires LibScriptableDisplayWidget-1.0")
local LibTimer = LibStub("LibScriptableDisplayTimer-1.0", true)
assert(LibTimer, MAJOR .. " requires LibScriptableDisplayTimer-1.0")
local Evaluator = LibStub("LibScriptableDisplayEvaluator-1.0", true)
assert(Evaluator, MAJOR .. " requires LibScriptableDisplayEvaluator-1.0")
local PluginUtils = LibStub("LibScriptableDisplayPluginUtils-1.0", true)
assert(PluginUtils, MAJOR .. " requires LibScriptableDisplayPluginUtils-1.0")
local L = LibStub("LibScriptableDisplayLocale-1.0", true)
assert(L, MAJOR .. " requires LibScriptableDisplayLocale-1.0")
L = L.L

if not WidgetColor.__index then WidgetColor.__index = WidgetColor end


local defaults = {
}
WidgetColor.defaults = defaults


local widgetType = {gesture=true}

local pool = setmetatable({}, {__mode = "v"})
local function new(...)
	local obj = table.remove(pool) or {}
	for i = 1, select("#", ...) do
		obj[i] = select(i, ...)
	end
	return obj
end

local function newDict(...)
	local obj = table.remove(pool) or {}
	for i = 1, select("#", ...), 2 do
		local key = select(i, ...)
		local val = select(i + 1, ...)
		obj[key] = val
	end
	return obj
end

local function del(obj)
	wipe(obj)
	tinsert(pool, obj)
end

local pool = setmetatable({}, {__mode = "k"})

local stopFunc

-- Create a new LibScriptableDisplayWidgetColor object
-- @name WidgetColor.New
-- @usage WidgetColor:New(visitor, name, config, errorLevel)
-- @param visitor An LibScriptableDisplayCore-1.0 object, or provide your own
-- @param name A name for the timer widget
-- @param config This timeer's parameters
-- @param errorLevel The errorLevel for this object
-- @param callback An optional callback function to be executed when the gesture is performed.
-- @param An optional timer. This should have a :Start() and :Stop().
-- @return A new LibScriptableDisplayWidgetColor widget
function WidgetColor:New(visitor, name, config, errorLevel, callback, timer) 
	assert(name, "WidgetColor requires a name.")
	assert(config, "Please provide the timer with a config")
	assert(config.expression, name .. ": Please provide the timer with an expression")
	
	local obj = next(pool)

	if obj then
		pool[obj] = nil
	else
		obj = {}
		obj.options = {}
	end
		
	setmetatable(obj, self)

	obj.widget = LibWidget:New(obj, visitor, MAJOR .. "." .. name, config, 0, 0, 0, widgetType, errorLevel)
	
	obj.expression = config.expression
		
	obj.timer = timer or LibTimer:New("WidgetColor.timer " .. obj.name, obj.update, obj.repeating, self.Update, obj, obj.errorLevel)
	
	return obj	
end


-- Delete a LibScriptableDisplayWidgetColor object
-- @name LibScriptableDisplayWidgetColor.Del
-- @usage LibScriptableDisplayWidgetColor:Del([marq])
-- @param marq A LibScriptableDisplayWidgetColor object
-- @return Nothing
function WidgetColor:Del()
	self:Stop()
	self.widget:Del()
	self.timer:Del()
	pool[timer] = true
end

-- Start a LibScriptableDisplayWidgetColor
-- @name LibScriptableDisplayWidgetColor.Start
-- @usage object:Start()
-- @return Nothing
function WidgetColor:Start()
	if self.update > 0 then
		self.timer:Start()
		self.active = true
	end
end

-- Stop a LibScriptableDisplayWidgetColor
-- @name LibScriptableDisplayWidgetColor.Stop
-- @usage object:Stop()
-- @return Nothing
function WidgetColor:Stop()
	self.timer:Stop()
	self.active = false
end

function WidgetColor:Update()

	selv.environment.self = self
	local r, g, b, a = Evaluator.ExecuteCode(obj.environment, MAJOR .. " startFunc", self.expression)
	selv.environment.self = nil
	
	if not r then return end
	
	if r and r > 1 then
		r = r / 255
	end
	
	if g and g > 1 then
		g = g / 255
	end
	
	if b and b > 1 then
		b = b / 255
	end
	
	if a and a > 1 then
		a = a / 255
	end
		
	if not g then
		if a then
			self.r, self.g, self.b, self.a = PluginUtils.Color2RGBA(r, true)
		else
			self.r, self.g, self.b =  PluginUtils.Color2RGBA(r)
		end
	else
		self.r, self.g, self.b, self.a = r, g, b, a
	end
end

-- Get an Ace3 option table. Plug this into a group type's args.
-- @param db The database table
-- @param callback Provide this if you want to execute the callback once an option is changed
-- @param data Some data to pass when executing the callback
-- @return An Ace3 options table -- `name.args = options`.
function WidgetColor:GetOptions(db, callback, data)
	local defaults = WidgetColor.defaults
	local options = {
		enabled = {
			name = "Enabled",
			desc = "Whether this timer is enabled or not",
			type = "toggle",
			get = function() return db.enabled end,
			set = function(info, v) 
				db.enabled = v
				db.enabledDirty = true
				if type(callback) == "function" then callback(data) end
			end,
			order = 5
		},
		update = {
			name = "Update Rate",
			desc = "Enter the timer's refresh rate",
			type = "input",
			pattern = "%d",
			get = function()
				return tostring(db.update or defaults.update)
				end,
			set = function(info, v)
				db.update = tonumber(v)
				db.updateDirty = true
				if type(callback) == "function" then callback(data) end
			end,
			order = 6
		},
		drawLayer = {
			name = L["Draw Layer"],
			type = "input",
			get = function()
				return db.drawLayer or defaults.drawLayer
			end,
			set = function(info, v)
				db.drawLayer = v
				db.drawLayerDirty = true
				if type(callback) == "function" then callback(data) end
			end,
			order = 9
		},
		startButton = {
			name = L["Start Button"],
			desc = L["A description of the mouse action that will trigger the start of the gesture recording. Possibly values are: LeftButtonUp, LeftButtonDown, RightButtonUp, RightButtonDown, MiddleButtonUp, MiddleButtonDown, Freehand"],
			type = "input",
			get = function()
				return db.startButton or defaults.startButton
			end,
			set = function(info, v)
				db.startButton = v
				db.startButtonDirty = true
				if type(callback) == "function" then callback(data) end
			end,
			order = 10
		},
		stopButton = {
			name = L["Stop Button"],
			desc = L["A description of the mouse action that will stop (finish) the gesture recording. Possible values are: LeftButtonUp, LeftButtonDown, RightButtonUp, RightButtonDown, MiddleButtonUp, MiddleButtonDown"],
			type = "input",
			get = function()
				return db.stopButton or defaults.stopButton
			end,
			set = function(info, v)
				db.stopButton = v
				db.stopButtonDirty = true
				if type(callback) == "function" then callback(data) end
			end,
			order = 11
		},
		nextButton = {
			name = L["Next Button"],
			desc = L["A description of the optional mouse action that will trigger nextFunc and start a new recording if defined. This can be used to simultaneously recording multiple gestures at once. Possible values are: LeftButtonUp, LeftButtonDown, RightButtonUp, RightButtonDown, MiddleButtonUp, MiddleButtonDown"],
			type = "input",
			get = function()
			
			end,
			set = function(info, v)
				db.nextButton = v
				db.nextButtonDirty = true
				if type(callback) == "function" then callback(data) end
			end,
			order = 12
		},
		cancelButton = {
			name = L["Cancel Button"],
			type = "input",
			desc = L["A description of the optional mouse action that will cancel the recording and call cancelFunc instead of stopFunc. Possible values are: LeftButtonUp, LeftButtonDown, RightButtonUp, RightButtonDown, MiddleButtonUp, MiddleButtonDown"],
			get = function()
				return db.cancelButton or defaults.cancelButton
			end,
			set = function(info, v)
				db.cancelButton = v
			end,
			order = 12
		},
		showTrail = {
			name = L["Show Trail"],
			type = "toggle",
			desc = L["If set to true, a cursor trail will be shown while the gesture takes place. (can be used for debugging and whatnot)"],
			get = function()
				return db.showTrail or defaults.showTrail
			end,
			set = function(info, v)
				db.showTrail = v
				db.showTrailDirty = true
				if type(callback) == "function" then
					callback(data)
				end
			end,
			order = 13
		},
		maxGestures = {
			name = L["Max Gestures"],
			desc = L["If startButton is set to Freehand, maxGestures can be used to set a hard cap on the number of gestures the user can make before the recording stops. Please note, that this will only work on linear gestures. If you need to record circular gestures, leave out this field and the library will sort it for you instead."],
			type = "input",
			pattern = "%d",
			get = function()
				return db.maxGestures or defaults.maxGestures
			end,
			set = function(info, v)
				db.maxGestures = tonumber(v)
				db.maxGesturesDirty = true
				if type(callback) == "function" then
					callback(data)
				end
			end,
			order = 14
		},
		repeating = {
			name = L["Repeating"],
			desc = L["Whether to keep repeating the recording or not"],
			type = "toggle",
			get = function()
				return db.repeating or defaults.repeating
			end,
			set = function(info, v)
				db.repeating = v
				db.repeatingDirty = true
				if type(callback) == "function" then
					callback(data)
				end
			end,
			order = 15
		},
		errorsAllowed = {
			name = L["Error Threshhold"],
			desc = L["The gesture will automatically fail if you make more than this many errors"],
			type = "input",
			pattern = "%d",
			get = function()
				return db.errorsAllowed or defaults.errorsAllowed
			end,
			set = function(info, v)
				db.errorsAllowed = v
				db.errorsAllowedDirty = true
				if type(callback) == "function" then
					callback(data)
				end
			end,
			order = 16
		},
		startFunc = {
			name = L["Start Function"],
			type = "input",
			desc = L["An optional function that will be called on the first start of a gesture recording. This can be used to initializing values and so on. startFunc will only be called at the start of a capture, and not after a nextButton has fired. Callback values: recorder, x1, y1, x2, y2 : The recorder object as well as the initial coordinates where the mouse triggered startFunc and the current cursor coordinates (which should be the same in this case)"],
			width = "full",
			multiline = true,
			get = function()
			
			end,
			set = function(info, v)
				db.startFunc = v
				db.startFuncDirty = true
				if type(callback) == "function" then
					callback(data)
				end
			end,
			order = 49
		},
		updateFunc = {
			name = L["Update Function"],
			desc = L["A function that will be called after each OnUpdate event on the recording frame. Callback values: recorder, x1, y1, x2, y2 : The recorder object as well as the initial coordinates where the mouse triggered startFunc (or the last called nextFunc) and the current cursor coordinates"],
			type = "input",
			width = "full",
			multiline = true,
			get = function()
				return db.updateFunc or defaults.updateFunc
			end,
			set = function(info, v)
				db.updateFunc = v
				db.updateFuncDirty = true
				if type(callback) == "function" then
					callback(data)
				end			
			end,
			order = 50
		},
		stopFunc = {
			name = L["Stop Function"],
			desc = L["A function that will be called when stopButton has triggered the end of a recording. Callback values: recorder, x1, y1, x2, y2 : The recorder object as well as the initial coordinates where the mouse triggered startFunc (or the last called nextFunc) and the current cursor coordinates"],
			type = "input",
			width = "full",
			multiline = true,
			get = function()
				return db.stopFunc or defaults.stopFunc
			end,
			set = function(info, v)
				db.stopFunc = v
				db.stopFuncDirty = true
				if type(callback) == "function" then
					callback(data)
				end			
			end,
			order = 51
		},
		
		nextFunc = {
			name = L["Next Function"],
			desc = L["A function that will be called when nextButton has triggered the stop of the current recording and immediate start of a new from the current position. Callback values: recorder, x1, y1, x2, y2 : The recorder object as well as the initial coordinates where the mouse triggered startFunc (or the last called nextFunc) and the current cursor coordinates."],
			type = "input",
			width = "full",
			multiline = true,
			get = function()
				return db.nextFunc
			end,
			set = function(info, v)
				db.nextFunc = v
				db.nextFuncDirty = true
				if type(callback) == "function" then
					callback(data)
				end
			end,
			order = 52
		},
		cancelFunc = {
			name = L["Cancel Function"],
			desc = L["A function that will be called when cancelButton has triggered the termination of a recording. Callback values: recorder, x1, y1, x2, y2 : The recorder object as well as the initial coordinates where the mouse triggered startFunc (or the last called nextFunc) and the current cursor coordinates"],
			type = "input",
			width = "full",
			multiline = true,
			get = function()
				return db.cancelFunc or defaults.cancelFunc
			end,
			set = function(info, v)
				db.cancelFunc = v
				db.cancelFuncDirty = true
				if type(callback) == "function" then
					callback(data)
				end			
			end,
			order = 53
		},
		tooltip	= {
			name = L["Tooltip"],
			desc = L["If set to either a string value or a table with multiple strings, a tooltip will be displayed next to the cursor when positioned over the recording frame. This can be used to provides tips on how to use the mouse gestures."],
			type = "input",
			width = "full",
			multiline = true,
			get = function()
				return db.tooltip or defaults.tooltip
			end,
			set = function(info, v)
				db.tooltip = v ~= "" and v or nil
				db.tooltipDirty = true
				if type(callback) == "function" then
					callback(data)
				end			
			end,
			order = 54
		},
		expression = {
			name = L["Expression"],
			desc = L["Enter this widget's expression"],
			type = "input",
			width = "full",
			multiline = true,
			get = function()
				return db.expression or defaults.expression
			end,
			set = function(info, v)
				db.expression = v
				db.expressionDirty = true
				if type(callback) == "function" then callback(data) end
			end,
			order = 55
		},
		
	}
	options.gestures = {
		name = L["Gestures"],
		type = "group",
		args = {
			add = {
				name = L["Add Gesture"],
				type = "execute",
				func = function()
					local gesture = {
						enabled = true,
						update = defaults.update,
						drawLayer = defaults.drawLayer,
						startButton = defaults.startButton, 
						stopButton = defaults.stopButton,
						cancelButton = defaults.cancelButton,
						showTrail = defaults.showTrail,
						maxGestures = defaults.maxGestures,					
					}
					tinsert(db.gestures, gesture)
					if type(callback) == "function" then
						callback(data)
					end
				end,
				order = 1
			},
		}
	}
	
	for i, gesture in ipairs(db.gestures) do
		options.gestures.args["Gesture" .. i] = {
			name = L["Gesture "] .. i,
			type = "group",
			order = i,
			args = {
				type = {
					name = L["Type"],
					type = "select",
					values = typesList,
					get = function()
						StarTip:Print(gesture.type)
						return gesture.type and types[gesture.type] or defaults.type
					end,
					set = function(info, v)
						gesture.type = types[v]
						gesture.typeDirty = true
						if type(callback) == "function" then callback(data) end
					end,
					order = 7
				},
				pattern = {
					name = L["Pattern"],
					type = "select",
					values = patternsList,
					get = function()
						return gesture.pattern and patterns[gesture.pattern] or defaults.pattern
					end,
					set = function(info, v)
						gesture.pattern = v
						gesture.patternDirty = true
						if type(callback) == "function" then callback(data) end
					end,
					order = 8			
				},
			}
		}	
	end
	return options
end
